/**
 * Copyright (C) 2011 Ovea <dev@ovea.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ovea.resource.web;

import com.ovea.resource.web.util.Gzip;
import com.ovea.resource.web.util.MimeTypes;
import com.ovea.resource.web.util.Resource;
import com.yahoo.platform.yui.compressor.CssCompressor;
import com.yahoo.platform.yui.compressor.JavaScriptCompressor;
import org.mozilla.javascript.ErrorReporter;
import org.mozilla.javascript.EvaluatorException;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;

final class VirtualResource {

    private static Logger LOGGER = Logger.getLogger(VirtualResource.class.getName());

    private final AtomicLong lastModificationTimeInSeconds = new AtomicLong(0);
    private final List<WebResource> webResources = new LinkedList<WebResource>();
    private final String virtualPath;
    private String contentType = MimeTypes.UNKNOWN;
    private long caching = -1;
    private boolean gzip;
    private byte[] content;
    private byte[] gzipContent;
    private boolean debug;

    VirtualResource(String virtualPath) {
        this.virtualPath = virtualPath;
        setContentType(MimeTypes.getContentTypeForPath(virtualPath));
    }

    public void setDebug(boolean debug) {
        this.debug = debug;
    }

    void setContentType(String contentType) {
        this.contentType = contentType;
    }

    @Override
    public String toString() {
        return virtualPath;
    }

    public String getVirtualPath() {
        return virtualPath;
    }

    void setCaching(long caching) {
        this.caching = Math.max(caching, 0);
    }

    void setGzip(boolean gzip) {
        this.gzip = gzip;
    }

    public WebResource addWebResource(Resource resource) {
        if (LOGGER.isLoggable(Level.FINE))
            LOGGER.fine(virtualPath + " : + " + resource);
        WebResource webResource = new WebResource(resource);
        webResources.add(webResource);
        return webResource;
    }

    public boolean needsCaching() {
        return caching > 0;
    }

    /**
     * last modif time in milliseconds, rouned to the second
     */
    public long getLastModificationTime() {
        return lastModificationTimeInSeconds.get();
    }

    /**
     * caching time in seconds
     */
    public long getCaching() {
        return caching;
    }

    public String getContentType() {
        return contentType;
    }

    public boolean isGzip() {
        return gzip;
    }

    public byte[] getContent() {
        return content != null ? content : buildContent();
    }

    public byte[] getGzipContent() {
        return gzipContent != null ? gzipContent : buildGzipContent();
    }

    public void preload() {
        content = buildContent();
        if (gzip)
            gzipContent = buildGzipContent();
    }

    private byte[] buildContent() {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        for (WebResource webResource : webResources) {
            if (LOGGER.isLoggable(Level.INFO))
                LOGGER.info("[" + virtualPath + "] Loading " + webResource + "...");
            try {
                baos.write(!debug && webResource.mustMinimize() ?
                    minimize(webResource) :
                    webResource.getData());
            } catch (Exception e) {
                throw new RuntimeException("Error building content of virtual resource " + virtualPath + " when processing resource " + webResource + " : " + e.getMessage(), e);
            }
        }
        modified();
        return baos.toByteArray();
    }

    private byte[] buildGzipContent() {
        return Gzip.compress(getContent());
    }

    private byte[] minimize(final WebResource webResource) {
        if ("text/css".equals(contentType)) {
            if (LOGGER.isLoggable(Level.INFO))
                LOGGER.info("[" + virtualPath + "] Minifying CSS " + webResource + "...");
            try {
                CssCompressor compressor = new CssCompressor(new StringReader(new String(webResource.getData())));
                StringWriter out = new StringWriter();
                compressor.compress(out, -1);
                return out.toString().getBytes();
            } catch (IOException e) {
                throw new IllegalStateException("[" + virtualPath + "] Unable to minimize CSS " + webResource + " : " + e.getMessage(), e);
            }
        } else if ("application/x-javascript".equals(contentType)) {
            if (LOGGER.isLoggable(Level.INFO))
                LOGGER.info("[" + virtualPath + "] Minifying Javascript " + webResource + "...");
            try {
                JavaScriptCompressor compressor = new JavaScriptCompressor(new StringReader(new String(webResource.getData())), new ErrorReporter() {
                    @Override
                    public void warning(String message, String sourceName, int line, String lineSource, int lineOffset) {
                        if (line < 0) {
                            if (LOGGER.isLoggable(Level.FINE))
                                LOGGER.fine("[" + virtualPath + "] Minifying Javascript " + webResource + " : " + message);
                        } else {
                            if (LOGGER.isLoggable(Level.FINE))
                                LOGGER.fine("[" + virtualPath + "] Minifying Javascript " + webResource + " : line " + line + ", offset " + lineOffset + " : " + message);
                        }
                    }

                    @Override
                    public void error(String message, String sourceName, int line, String lineSource, int lineOffset) {
                        if (line < 0) {
                            LOGGER.severe("[" + virtualPath + "] Minimizing Javascript " + webResource + " : " + message);
                        } else {
                            LOGGER.severe("[" + virtualPath + "] Minimizing Javascript " + webResource + " : " + line + ':' + lineOffset + ':' + message);
                        }
                    }

                    @Override
                    public EvaluatorException runtimeError(String message, String sourceName, int line, String lineSource, int lineOffset) {
                        error(message, sourceName, line, lineSource, lineOffset);
                        return new EvaluatorException(message);
                    }
                });
                StringWriter out = new StringWriter();
                compressor.compress(out, -1, false, true, false, false);
                return out.toString().getBytes();
            } catch (IOException e) {
                throw new IllegalStateException("[" + virtualPath + "] Unable to minimize Javascript " + webResource + " : " + e.getMessage(), e);
            }
        }
        return webResource.getData();
    }

    private void modified() {
        long time = System.currentTimeMillis();
        time = time - (time % 1000);
        lastModificationTimeInSeconds.set(time);
    }

}