/**
 * Copyright (C) 2011 Ovea <dev@ovea.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ovea.resource.web;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;

public class WebResourceServlet extends HttpServlet {

    public static final String DESCRIPTOR = "descriptor";

    private XmlBasedWebResourceProvider provider;
    private boolean debug;

    public boolean isDebug() {
        return debug;
    }

    public void setDebug(boolean debug) {
        this.debug = debug;
    }

    @Override
    public void init() throws ServletException {
        String debug = getServletConfig().getInitParameter("debug");
        if(debug != null)
            setDebug(Boolean.valueOf(debug));
        String descriptor = getServletConfig().getInitParameter(DESCRIPTOR);
        if (descriptor.startsWith("/"))
            descriptor = descriptor.substring(1);
        if (descriptor == null)
            throw new IllegalStateException("Missing init parameter: " + DESCRIPTOR);
        provider = new XmlBasedWebResourceProvider(getServletContext(), descriptor);
        provider.setDebug(isDebug());
        if (!isDebug())
            provider.get().preload();
    }

    @Override
    protected void doGet(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
        String path = request.getRequestURI();
        if (path.startsWith(request.getContextPath()))
            path = path.substring(request.getContextPath().length());
        if (path.startsWith(request.getServletPath()))
            path = path.substring(request.getServletPath().length());
        if (path != null) {
            int pos = path.indexOf(";jsessionid");
            if (pos != -1)
                path = path.substring(0, pos);
            VirtualResources virtualResources = provider.get();
            if (virtualResources.has(path)) {
                VirtualResource resource = virtualResources.get(path);
                if (!isDebug() && resource.needsCaching()) {
                    long clientLastModified = request.getDateHeader("If-Modified-Since");
                    long lmt = resource.getLastModificationTime();
                    if (clientLastModified != -1 && clientLastModified >= lmt) {
                        response.setStatus(HttpServletResponse.SC_NOT_MODIFIED);
                        return;
                    }
                    response.setDateHeader("Last-Modified", lmt);
                    response.setHeader("Cache-Control", "max-age=" + resource.getCaching() + ", must-revalidate");
                    response.setDateHeader("Expires", lmt + resource.getCaching() * 1000);
                } else {
                    response.setHeader("Cache-Control", "no-store, no-cache, must-revalidate, max-age=0");
                    response.setHeader("Pragma", "no-cache");
                }
                response.setContentType(resource.getContentType());
                byte[] data = resource.getContent();
                if (resource.isGzip() && acceptsGZipEncoding(request)) {
                    byte[] compressed = resource.getGzipContent();
                    response.setHeader("Content-Encoding", "gzip");
                    response.setContentLength(compressed.length);
                    response.getOutputStream().write(compressed);
                } else {
                    response.setContentLength(data.length);
                    response.getOutputStream().write(data);
                }
                response.setStatus(HttpServletResponse.SC_OK);
                response.flushBuffer();
                return;
            }
        }
        response.sendError(HttpServletResponse.SC_NOT_FOUND);
    }

    private static boolean acceptsGZipEncoding(HttpServletRequest request) {
        String acceptEncoding = request.getHeader("Accept-Encoding");
        return acceptEncoding != null && acceptEncoding.indexOf("gzip") != -1;
    }

}